package   cohy.estDip;

import 	  sep.operator.*;
import    jam.vector.*;
import 	  rsf.vector.*;
import   cohy.gradient.*;
import   conv.convi.*;
import   sep.io.ArgumentKeeper;

/**
 * Computes the crossproduct <tt>({\bf p \times \nabla) f</tt>. 
 * If <tt>f<tt> is a plane-wave and <tt>p</tt> its dip, the operator 
 * will return a null vector.<BR> 
 * In 2D the operator is simply the partial differential operator 
 * <tt>p_y \partial_x + p_x \partial_y</tt> and returns a scalar 
 * for each input point. 
 * In 3D the operator yields a 3 component vector for each input 
 * point.<BR> 
 * 
 * In the 3-D case, I currently compute 
 * a single output component at a given run. This avoids memory problems 
 * and makes the output handling easier. In 3D cases, the operator computes 
 * actually 
 * <tt>(\kronecker_{i-1},\kronecker_{i-2},\kronecker_{i-3}) 
 * ({\bf p \times \nabla) f</tt>.<BR>  
 *
 * Maybe a better design to hand in the dip and to make this the linear 
 * part of a SemiLinear operator. I also should pack the 6 dimensional 
 * output into a single vector. But I don't think the result will warrant 
 * the effort. 
 *
 * Keeping the 2-D version and the 3-D version in the same file seems to 
 * create a mess. I probably should split them. 
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 */
public class DipXGrad extends SepOperator{
  final static int XCOMPONENT = 0;
  final static int YCOMPONENT = 1;
  final static int ZCOMPONENT = 2;
  int wantedResult = DipXGrad.ZCOMPONENT;

  DipEstimator dipy; // could create on the fly: clutter vs efficiency
  Vector       dips; // could create on the fly: clutter vs efficient?

  public DipXGrad(RsfSpace iputSpace) {
    this(iputSpace, DipXGrad.getOutputRsfSpace(iputSpace));
  }
  public DipXGrad(RsfSpace iputSpace, RsfSpace oputSpace) {
    super(iputSpace, oputSpace);
    int dim   = iputSpace.getDimension();
    if (dim < 2 || dim > 3) System.err.println("Error: not implemented ..");
    this.dipy = new DipEstimator(iputSpace);
    this.dips = dipy.getRange().newMember();
  }
  protected DipXGrad(DipXGrad src) {
    super(src);
    dipy = new DipEstimator((RsfSpace) src.getDomain()); 
    dips = dipy.getRange().newMember(); // should this copy the values? 
  }
  public jam.JamObject copy() { return new DipXGrad(this);}

  /**
   * This should be using the IcaiFactory class once that class implements
   * the OperatorFactory interface. Until then this class is a hack. 
   * I should probably call this createRange(). 
   */
  public static RsfSpace getOutputRsfSpace(RsfSpace iputSpace) {
    int n = iputSpace.getDimension();
    Axis[] oputAxes = new Axis[n];
    for (int i = 0 ; i < n ; i++) {
      oputAxes[i]          = new Axis(iputSpace.getAxes()[i]);
      oputAxes[i].nsamples =          iputSpace.getNSamples(i) -1;
    }
    return new RsfSpace(oputAxes); 
  }
  /**
   * I should call this createDomain(). More comments see getOutputRsfSpace.
   */
  public static RsfSpace getInputRsfSpace(RsfSpace oputSpace) {
    int n = oputSpace.getDimension();
    Axis[] iputAxes = new Axis[n];
    for (int i = 0 ; i < n ; i++) {
      iputAxes[i]          = new Axis(oputSpace.getAxes()[i]);
      iputAxes[i].nsamples =          oputSpace.getNSamples(i) +1;
    }
    return new RsfSpace(iputAxes); 
  }
  public void setResult(int comp) { 
    if (comp >= 0 && comp <= 2) 
      wantedResult = comp;
    else 
      System.err.println("Error: dimension not implemented");
  }
  public void set3DToX() { wantedResult = DipXGrad.XCOMPONENT; }
  public void set3DToY() { wantedResult = DipXGrad.YCOMPONENT; }
  public void set3DToZ() { wantedResult = DipXGrad.ZCOMPONENT; }

  protected void apply(boolean add, Vector iput, Vector oput) {
    int dim  = ((Rsf) iput).getDimension();
    switch (dim) {
    case 2: apply2d(add, iput, oput); break;
    case 3: apply3d(add, iput, oput); break;
    default:
      System.err.println("Error: dimensions not implemented yet."); 
      break;
    }
  }

  void apply2d(boolean add, Vector iput, Vector oput) {
    dipy.image(iput, dips); 
    int[] idx = {DipEstimator.DX}; float dx = ((Rsf)dips).getDatum(idx);
    int[] idy = {DipEstimator.DY}; float dy = ((Rsf)dips).getDatum(idy);

    // create the convolution operator. 
    Rsf ff = PartDiff2d.getDerivativeDx();
    InternalConvolution diff = 
      new InternalConvolution( InternalConvolution.kInput , 
			       (RsfSpace) iput.getSpace(),
			       InternalConvolution.kFilter,
			       (RsfSpace) ff.getSpace());

    //Rsf fd   = PartDiff2d.getDerivative(1.f, -dx/dy);
    Rsf fd   = PartDiff2d.getDerivative(dy, -dx);
    // apply the differential 
    diff.imageBinary(iput, fd, oput);
  }

  void apply3d(boolean add, Vector iput, Vector oput) {
    dipy.image(iput, dips); 
    int[] idx = {DipEstimator.DX}; float dx = ((Rsf)dips).getDatum(idx);
    int[] idy = {DipEstimator.DY}; float dy = ((Rsf)dips).getDatum(idy);
    int[] idz = {DipEstimator.DZ}; float dz = ((Rsf)dips).getDatum(idz);

    // create the convolution operator. 
    Rsf ff = PartDiff3d.getDerivativeDx();
    InternalConvolution diff = 
      new InternalConvolution( InternalConvolution.kInput , 
			       (RsfSpace) iput.getSpace(),
			       InternalConvolution.kFilter,
			       (RsfSpace) ff.getSpace());

    Rsf fd; 
    switch (wantedResult) {
    case DipXGrad.XCOMPONENT:
      //System.err.println("I do X");
      fd   = PartDiff3d.getDerivative(0.f, -dz,  dy);
      break;
    case DipXGrad.YCOMPONENT:
      //System.err.println("I do Y");
      fd   = PartDiff3d.getDerivative( dz, 0.f, -dx);
      break;
    case DipXGrad.ZCOMPONENT: default:
      //System.err.println("I do Z");
      fd   = PartDiff3d.getDerivative(-dy,  dx, 0.f);
      break;
    }
    //System.err.println("fd: " + fd); 
    // apply the differential 
    diff.imageBinary(iput, fd, oput);
  }
}


