package  geom.proj;
import   geom.simple.*;
import   jam.vector.*;
import   sep.operator.*;
import   rsf.vector.*;
import   sep.io.*;
import   conv.convc.*;
import   rsf.operator.*;

/** 
 * This class generates a synthetic Rsf with (random) plane waves.<BR> 
 *
 * This computes f(n*x) where                        <BR> 
 *                 f is a scalar (random) function.  <BR> 
 *                 x is a multidimensional variable. <BR> 
 *                 n is a stepout vector that determines the phase of f.<BR> 
 * The input and output vector are identical. 
 *
 * <BR><!--#include file="./source.inc"--><BR>
 * @author Matthias Schwab
 */
public class PlaneWave extends SepOperator {
  private RsfSpace       range = null; 
  private TraceGenerator tgtor = null; 
  private Ccai           tfilt = null;
  private Projection     pctor = null; 
  private Ccai           rfilt = null;

  public PlaneWave(RsfSpace planeWaveRange) {
    this(new TraceGenerator(Projection.createRange(planeWaveRange)),
	 new Projection                           (planeWaveRange) ,
	                                           planeWaveRange);
  }

  public PlaneWave(TraceGenerator tgtor, 
	    Projection     pctor, RsfSpace range) {
    super(range, range); 
    this.range = range; 
    this.tgtor = tgtor; 
    this.pctor = pctor; 
  }
  /** 
   * Constructor accepts TraceGenerator, TraceFilter, Projector, DataFilter. 
   * Use Factory to ensure that all arguments are consistent. 
   */
  public PlaneWave(TraceGenerator tgtor, Ccai tfilt, 
	    Projection     pctor, Ccai rfilt, RsfSpace range) {
    super(range, range); 
    this.range = range; 
    this.tgtor = tgtor; 
    this.tfilt = tfilt; 
    this.pctor = pctor; 
    this.rfilt = rfilt; 
  }
  PlaneWave(PlaneWave src) {
    this.range = (RsfSpace)   src.range.copy();
    this.tgtor =              src.tgtor.copy(); 
    this.tfilt = (Ccai)       src.tfilt.copy();
    this.pctor = (Projection) src.pctor.copy(); 
    this.rfilt = (Ccai)       src.rfilt.copy(); 
  }
  public jam.JamObject copy() { return new PlaneWave(this); }

  public void setTraceType  (int     type) { tgtor.setTraceType(type); } 
  public void setTraceSeed  (long    seed) { tgtor.setSeed(     seed); }
  /** Sets normal. steps = n = (n_x,n_y,n_z) of f(<BF>x n</BF>) */
  public void setStepout    (float[] steps){ pctor.setStepout( steps); }
  /** 
   * Sets normal. 
   * Angle of normal vector (polar     coords).
   * Angle of normal vector (spherical coords).
   * @see rsf.operator.Projection  
   */ 
  public void setNormal    (double[] alpha){ pctor.setNormal(alpha); }
  public void setTraceFilter(Rsf     filt) { 
    CcaiFactory fact = new CcaiFactory(filt);
    tfilt     = fact.getCcaiFromInputSpace(Projection.createRange(range));
  } 
  public void setRangeFilter(Rsf     filt) { 
    CcaiFactory fact = new CcaiFactory(filt);
    rfilt     = fact.getCcaiFromInputSpace(range);
  } 
  
  protected void apply(boolean add, Vector dom, Vector ran) { 
    if (!add) ran.zero(); 

    Rsf tmpTrc = (Rsf) pctor.getRange().newMember();
    Rsf tmpRan = (Rsf) pctor.getDomain().newMember();

    if (tfilt != null) { 
      tfilt.addImage(tgtor.createTrace(), tmpTrc);
    } 
    else {
      tmpTrc =       tgtor.createTrace(); 
    }
    if (rfilt != null) { 
      pctor.adjoint().addImage(tmpTrc, tmpRan); 
      rfilt.addImage(          tmpRan,    ran);
    } 
    else {
      pctor.adjoint().addImage(tmpTrc,    ran); 
    }               
  } 

  /**
   * @param int     boxlength= 1; 
   * @param int[]   nsamples = {100,20,10}
   * @param float[] stepout1 = {1.f,  1.f, 1.f}
   * @param float[] stepout2 = {1.f, -1.f, 1.f}
   */
  public static void main(String[] args) {
    ArgumentKeeper   cmd = new ArgumentKeeper(args);
    Rsf data = RsfFactory.newRsf(cmd.getIntArray("nsamples","100,20,10"));
    PlaneWave gtor = new PlaneWave((RsfSpace) data.getSpace());
    float[] stepout;
    String  cmdName = new String("stepout0"); 
    for (int i=0; cmd.containsKey(cmdName); i++) {
      stepout = cmd.getFloatArray(cmdName);
      if (stepout.length != data.getDimension()) 
	System.err.println("Error: dimensional mismatch"); 
      gtor.setStepout(stepout); 
      gtor.addImage(data,data);      
      cmdName = new String("stepout" + i); 
    }
    if (cmd.containsKey("boxlength")) 
      gtor.setTraceFilter(rsf.util.Box.newRsf(cmd.getInt("boxlength"))); 
    if (cmd.containsKey("out")) data.write(cmd.getString("out")); 
    else                        data.write(System.out); 
  }

  public String toString() {
    return new String("Range:           \n" + range + 
		      "Trace Generator: \n" + tgtor + 
		      "Trace Filter:    \n" + tfilt + 
		      "Projector:       \n" + pctor + 
		      "Data  Filter:    \n" + rfilt ); 
  }
} 

