package  miss.bisolver;

import    jam.vector.*;
import    jam.operator.*;
import    rsf.vector.*;
import    rsf.solver.Mask;
import    rsf.solver.SepMask;
import  juice.operator.*;

/**
 * Sets up the linear missing data bilinear equations:         <BR>
 * Given: a bilinear operator B(.,.) and known data k1 and k2. <BR>
 * What we want to find is missing data x1 and x2 such that:   <BR>
 * B(k1 + x1, k2 + x2) ~= 0                                    <BR>
 * <BR>
 * Since B is bilinear, that is equivalent to:                 <BR>
 * B(x1,k2) + B(k1,x2) ~= -B(k1,k2) - B(x1,x2)                 <BR>
 * <BR>
 * Dropping the nonlinear term on the right side, we have a linear problem:<BR>
 * Ax = b                                                      <BR>
 * where x = [x1 x2]' , b = -B(k1,k2)                          <BR>
 * and A = [B(.,k2)M1 B(k1,.)M2]                               <BR>
 * Here M1 and M2 are free masks for x1 and x2                 <BR>
 * <BR>
 * For example, in SEP terminology, if B is transient convolution, <BR>
 * A = [tcai(known-filter)(free data mask) tcaf(known-data)(free filter mask)]
 * <BR>
 * x = [missing-data missing-filter-coefficients]'
 * <BR><BR>
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author    Joel Schroeder
 */
public class BiLinMissDataEquations {
  BlockLinearOperatorAdjoint A;
  Vector                     b;
  Mask                       M1, M2;

  public BiLinMissDataEquations(BiLinearOperatorAdjoint B,
				Rsf known1, Rsf known2) {
    this(B,
	 known1, new SepMask(known1, true),
	 known2, new SepMask(known2, true));
  }

  public BiLinMissDataEquations(BiLinearOperatorAdjoint B,
				Vector known1, Mask freeMask1,
				Vector known2, Mask freeMask2) {
    M1 = freeMask1;
    M2 = freeMask2;
    A = new BlockLinOpAdj(B.getDomain(), 2, B.getRange(), 1);
    A.setOp(0,0, new CompoundLinOpAdj(M1,
				      (hasAdjoint) B.getOp2(known2)));
    A.setOp(0,1, new CompoundLinOpAdj(M2,
				      (hasAdjoint) B.getOp1(known1)));
    b = B.imageBinary(known1, known2);
    b.neg();
  }

  hasAdjoint getA() 	    { return  A; }
  Vector     getb() 	    { return  b; }
  Mask       getFreeMask1() { return M1; }
  Mask       getFreeMask2() { return M2; }
}
